@@ -16,7 +16,7 @@
 
 use crate::{Arc, String, Vec};
 pub use snarkvm_algorithms::fft::DensePolynomial as Polynomial;
-use snarkvm_fields::Field;
+use snarkvm_fields::{ConstraintFieldError, Field, ToConstraintField};
 use snarkvm_utilities::{
     bytes::{FromBytes, ToBytes},
     error as error_fn,
@@ -63,6 +63,13 @@ pub trait PCVerifierKey: CanonicalSerialize + CanonicalDeserialize + Clone + Deb
     fn supported_degree(&self) -> usize;
 }
 
+/// Defines the minimal interface of prepared verifier keys for any polynomial
+/// commitment scheme.
+pub trait PCPreparedVerifierKey<Unprepared: PCVerifierKey> {
+    /// prepare
+    fn prepare(vk: &Unprepared) -> Self;
+}
+
 /// Defines the minimal interface of commitments for any polynomial
 /// commitment scheme.
 pub trait PCCommitment: CanonicalDeserialize + CanonicalSerialize + Clone + Debug + ToBytes {
@@ -76,6 +83,13 @@ pub trait PCCommitment: CanonicalDeserialize + CanonicalSerialize + Clone + Debu
     fn is_in_correct_subgroup_assuming_on_curve(&self) -> bool;
 }
 
+/// Defines the minimal interface of prepared commitments for any polynomial
+/// commitment scheme.
+pub trait PCPreparedCommitment<Unprepared: PCCommitment>: Clone {
+    /// Prepare
+    fn prepare(commitment: &Unprepared) -> Self;
+}
+
 /// Defines the minimal interface of commitment randomness for any polynomial
 /// commitment scheme.
 pub trait PCRandomness: CanonicalSerialize + CanonicalDeserialize + Clone {
@@ -114,22 +128,6 @@ impl<F: Field> core::ops::Deref for LabeledPolynomial<F> {
 
 impl<F: Field> LabeledPolynomial<F> {
     /// Construct a new labeled polynomial by consuming `polynomial`.
-    pub fn new_owned(
-        label: PolynomialLabel,
-        polynomial: Polynomial<F>,
-        degree_bound: Option<usize>,
-        hiding_bound: Option<usize>,
-    ) -> Self {
-        Self {
-            label,
-            polynomial: Arc::new(polynomial),
-            degree_bound,
-
-            hiding_bound,
-        }
-    }
-
-    /// Construct a new labeled polynomial.
     pub fn new(
         label: PolynomialLabel,
         polynomial: Polynomial<F>,
@@ -183,6 +181,12 @@ pub struct LabeledCommitment<C: PCCommitment> {
     degree_bound: Option<usize>,
 }
 
+impl<F: Field, C: PCCommitment + ToConstraintField<F>> ToConstraintField<F> for LabeledCommitment<C> {
+    fn to_field_elements(&self) -> Result<Vec<F>, ConstraintFieldError> {
+        self.commitment.to_field_elements()
+    }
+}
+
 // NOTE: Serializing the LabeledCommitments struct is done by serializing
 // _WITHOUT_ the labels or the degree bound. Deserialization is _NOT_ supported,
 // and you should construct the struct via the `LabeledCommitment::new` method after
@@ -285,7 +289,7 @@ pub struct LinearCombination<F> {
     /// The label.
     pub label: String,
     /// The linear combination of `(coeff, poly_label)` pairs.
-    terms: Vec<(F, LCTerm)>,
+    pub terms: Vec<(F, LCTerm)>,
 }
 
 impl<F: Field> LinearCombination<F> {
